import string

from logging import getLogger
log = getLogger('seantis.dir.events')

from datetime import timedelta
from dateutil.parser import parse

from lxml import objectify
from urllib import urlopen

from zope.component import getMultiAdapter
from zope.interface import Interface, Attribute


class IGuidleConfig(Interface):
    """ Multi Adapter Interface for context/request.
    See vbeo.seantis.dir.events.guidle for an example.

    """

    url = Attribute("Guidle URL to download from")

    classification = Attribute(
        "Name of classification to use (guidle:classification name)"
    )

    tagmap = Attribute(
        "Dictionary mapping tags to categories (guidle:classification tags)"
    )

    def on_event(root, offer, event):
        """ Called before an event is yielded to the source. Root is the
        xml root, offer is the xml offer element and event is the dictionary
        that is returned to the source handler.

        Since an offer may be split in multiple events this method may be
        called multiple times for each offer. The difference will be the
        event's dates.

        This function may then alter the event.

        """


def generate_recurrence(date):
    try:
        weekdays = list(date.weekdays.iterchildren())
        return "RRULE:FREQ=WEEKLY;BYDAY=%s" % ','.join(
            [d.text.upper() for d in weekdays]
        )
    except AttributeError:
        return ""


def limit_recurrence(recurrence, end):
    if not end or not recurrence:
        return recurrence
    else:
        return recurrence + ';UNTIL=%s' % end.strftime('%Y%m%dT%H%MZ')


def get_dates(date):

    def child(name):
        try:
            return getattr(date, name).text
        except AttributeError:
            return None

    return (
        child('startDate'), child('endDate'),
        child('startTime'), child('endTime')
    )


def events(offer):
    """ Generates one event for each date in the guidle input. """

    dates = [d for d in offer.schedules.iterchildren()]
    if not dates:
        log.error(
            'offer with id %s has no dates, skipping' % offer.attrib['id']
        )

        raise StopIteration

    for date in dates:

        start, end, start_time, end_time = get_dates(date)

        event = {}

        event['id'] = offer.attrib['id']
        event['recurrence'] = generate_recurrence(date)
        event['start'] = parse(start)

        # the end date of recurring events seems to be the date of the
        # last occurrence, which makes sense since guidle only seems
        # to support daily occurences by weekday
        if end and not event['recurrence']:
            event['end'] = parse(end)
        else:
            # if a recurrence exists it needs to be limited by the end date
            if event['recurrence']:
                until = parse(end) + timedelta(days=1)
                event['recurrence'] = limit_recurrence(
                    event['recurrence'], until
                )

            event['end'] = event['start']

        if not any((start_time, end_time)):
            event['whole_day'] = True
        else:
            event['whole_day'] = False

        if start_time:
            event['start'] = parse(start_time, default=event['start'])

        if end_time:
            event['end'] = parse(end_time, default=event['end'])

        # if the event ends before it starts it means that we have a range
        # like this: 20:00 - 00:30, so the event really ends the next day
        if event['end'] < event['start']:
            event['end'] += timedelta(days=1)

        yield event


def categories_by_tags(classification, tagmap):
    categories = set()

    for tagname in (c.attrib['name'] for c in classification.iterchildren()):
        for key in tagmap:
            if tagname.startswith(key):
                categories.add(tagmap[key])

    return categories


def copy(event, node, expression):
    lines = map(string.strip, expression.split('\n'))

    for line in lines:
        if not line:
            continue

        key, children = map(string.strip, line.split('<-'))

        for child in map(string.strip, children.split(',')):
            if not hasattr(node, child):
                continue

            if key in event:
                event[key] += '\n\n' + getattr(node, child).text
            else:
                event[key] = getattr(node, child).text


def fetch_events(context, request):

    config = getMultiAdapter((context, request), IGuidleConfig)

    xml = urlopen(config.url).read()
    root = objectify.fromstring(xml)

    offers = root.xpath(
        '*//guidle:offer',
        namespaces={'guidle': 'http://www.guidle.com'}
    )

    for offer in offers:

        last_update_of_offer = parse(
            offer.lastUpdateDate.text
        ).replace(
            microsecond=0
        )

        assert last_update_of_offer, """"
            offers should all have an update timestamp
        """

        for e in events(offer):

            e['fetch_id'] = config.url
            e['last_update'] = last_update_of_offer

            # so far all guidle events seem to be in this region
            e['timezone'] = 'Europe/Zurich'
            e['source_id'] = offer.attrib['id']

            # basic information
            copy(e, offer.offerDetail, """
                title               <- title
                short_description   <- shortDescription
                long_description    <- longDescription, openingHours
                prices              <- priceInformation
                event_url           <- externalLink
                location_url        <- homepage
                registration        <- ticketingUrl
            """)

            # address
            copy(e, offer.address, """
                locality    <- company
                street      <- street
                zipcode     <- zip
                town        <- city
                latitude    <- latitude
                longitude   <- longitude
            """)

            # contact
            copy(e, offer.contact, """
                organizer       <- company
                contact_name    <- name
                contact_email   <- email
                contact_phone   <- telephone_1
            """)

            # categories
            for classification in offer.classifications.iterchildren():

                if classification.attrib['name'] != config.classification:
                    continue

                e['cat1'] = categories_by_tags(classification, config.tagmap)
                e['cat2'] = set((e['town'],))

            # image (download later)
            try:
                for image in list(offer.offerDetail.images.iterchildren())[:1]:
                    copy(e, image, "image <- url")
            except AttributeError:
                pass

            # attachments (download later)
            try:
                attachments = list(
                    offer.offerDetail.attachments.iterchildren()
                )[:2]
                for ix, attachment in enumerate(attachments):
                    copy(e, attachment, "attachment_%i <- url" % (ix + 1))
            except AttributeError:
                pass

            config.on_event(root, offer, e)

            yield e



from seantis.dir.events.tests import IntegrationTestCase
from Products.CMFCore.WorkflowCore import WorkflowException


class TestWorkflow(IntegrationTestCase):

    def setUp(self):
        super(TestWorkflow, self).setUp()
        # login to gain the right to create events
        self.login_testuser()

    def tearDown(self):
        super(TestWorkflow, self).tearDown()
        self.logout()

    def test_events_workflow(self):

        # ensure that the initial state is preview
        event = self.create_event()
        self.assertEqual(event.state, 'preview')

        # go through the states
        event.do_action('submit')
        self.assertEqual(event.state, 'submitted')

        event.do_action('publish')
        self.assertEqual(event.state, 'published')

        event.do_action('archive')
        self.assertEqual(event.state, 'archived')

        event.do_action('publish')
        self.assertEqual(event.state, 'published')

        # try some impossible changes
        event = self.create_event()
        action = lambda a: event.do_action(a)

        self.assertRaises(WorkflowException, action, 'publish')
        self.assertRaises(WorkflowException, action, 'archive')

        event.do_action('submit')

        self.assertRaises(WorkflowException, action, 'archive')

        event.do_action('publish')

        self.assertRaises(WorkflowException, action, 'submit')

        event.do_action('archive')

        self.assertRaises(WorkflowException, action, 'submit')

        # try denying an event (skipping the publication)
        event = self.create_event()

        event.do_action('submit')
        self.assertEqual(event.state, 'submitted')

        event.do_action('deny')
        self.assertEqual(event.state, 'archived')

    def test_access_guard(self):
        actions = ['submit', 'publish', 'deny', 'archive']

        # the default access guard has no limits
        event = self.create_event()

        for action in actions:
            self.assertTrue(event.allow_action(action))

        # use a custom guard to ensure that the guards are actually
        # set in the definition xml file
        event = self.create_event()
        action = lambda a: event.do_action(a)

        event.allow_action = lambda a: False

        self.assertRaises(WorkflowException, action, 'submit')

        event.allow_action = lambda a: True

        action('submit')



import logging
log = logging.getLogger('seantis.dir.events')

from five import grok
from pytz import timezone

from datetime import datetime
from datetime import date as datetime_date
from urllib import urlopen
from icalendar import Calendar
from plone.dexterity.utils import createContent, addContentToContainer
from zope.component import queryAdapter
from zope.component.hooks import getSite
from Products.CMFPlone.PloneBatch import Batch

from seantis.dir.base import directory
from seantis.dir.base import session
from seantis.dir.base.utils import cached_property

from seantis.dir.events.unrestricted import execute_under_special_role
from seantis.dir.events.interfaces import (
    IEventsDirectory, IActionGuard
)

from seantis.dir.events.recurrence import grouped_occurrences
from seantis.dir.events import dates
from seantis.dir.events import utils
from seantis.dir.events import maintenance
from seantis.dir.events import _

from AccessControl import getSecurityManager
from Products.CMFCore import permissions

from seantis.dir.events import pages


class EventsDirectory(directory.Directory, pages.CustomPageHook):

    def labels(self):
        return dict(cat1=_(u'What'), cat2=_(u'Where'))

    def used_categories(self):
        return ('cat1', 'cat2')

    def unused_categories(self):
        return ('cat3', 'cat4')

    def allow_action(self, action, item_brain):
        """ Return true if the given action is allowed. This is not a
        wrapper for the transition guards of the event workflow. Instead
        it is called *by* the transition guards.

        This allows a number of people to work together on an event website
        with every person having its own group of events which he or she is
        responsible for.

        There's no actual implementation of that in seantis.dir.events
        but client specific packages like izug.seantis.dir.events may
        use a custom adapter to implement such a thing.
        """
        guard = queryAdapter(self, IActionGuard)

        if guard:
            return guard.allow_action(action, item_brain)
        else:
            return True


class ExtendedDirectoryViewlet(grok.Viewlet, pages.CustomDirectory):
    grok.context(IEventsDirectory)
    grok.name('seantis.dir.events.directory.detail')
    grok.require('zope2.View')
    grok.viewletmanager(directory.DirectoryViewletManager)

    template = grok.PageTemplateFile('templates/directorydetail.pt')

    def __init__(self, *args, **kwargs):
        super(ExtendedDirectoryViewlet, self).__init__(*args, **kwargs)
        self.context = self.custom_directory


class EventsDirectoryIndexView(grok.View, directory.DirectoryCatalogMixin):

    grok.name('eventindex')
    grok.context(IEventsDirectory)
    grok.require('cmf.ManagePortal')

    template = None

    def render(self):

        self.request.response.setHeader("Content-type", "text/plain")

        if 'rebuild' in self.request:
            log.info('rebuilding ZCatalog')
            self.catalog.catalog.clearFindAndRebuild()

        if 'reindex' in self.request:
            log.info('reindexing event indices')
            self.catalog.reindex()

        result = []
        for name, index in self.catalog.indices.items():

            result.append(name)
            result.append('-' * len(name))
            result.append('')

            for ix, identity in enumerate(index.index):
                result.append('%i -> %s' % (ix, identity))

            result.append('')

            dateindex = index.get_metadata('dateindex')

            if dateindex:
                result.append('-> dateindex')

                for date in sorted(dateindex):
                    result.append('%s -> %s' % (
                        date.strftime('%y.%m.%d'), dateindex[date])
                    )

        return '\n'.join(result)


class EventsDirectoryView(directory.View, pages.CustomDirectory):

    grok.name('view')
    grok.context(IEventsDirectory)
    grok.require('zope2.View')

    template = None
    _template = grok.PageTemplateFile('templates/directory.pt')

    @property
    def title(self):
        return self.custom_directory.title

    @property
    def is_ical_export(self):
        """ Returns true if the current request is an ical request. """
        return self.request.get('type') == 'ical'

    @property
    def is_json_export(self):
        """ Returns true if the current request is an json request. """
        return self.request.get('type') == 'json'

    def get_last_daterange(self):
        """ Returns the last selected daterange. """
        return session.get_session(self.context, 'daterange') \
            or dates.default_daterange

    def set_last_daterange(self, method):
        """ Store the last selected daterange on the session. """
        session.set_session(self.context, 'daterange', method)

    def get_last_state(self):
        """ Returns the last selected event state. """
        return session.get_session(self.context, 'state') or 'published'

    def set_last_state(self, method):
        """ Store the last selected event state on the session. """
        session.set_session(self.context, 'state', method)

    @property
    def no_events_helptext(self):
        if 'published' == self.catalog.state:
            return _(u'No events for the current daterange')
        else:
            return _(u'No events for the current state')

    @property
    def selected_daterange(self):
        return self.catalog.daterange

    @property
    def dateranges(self):
        return dates.methods

    def daterange_url(self, method):
        return self.directory.absolute_url() + '?range=' + method

    @property
    def has_results(self):
        return len(self.batch) > 0

    def render(self):
        """ Renders the ical/json if asked, or the usual template. """
        search = 'search' in self.request.keys() \
            and self.request.get('searchtext') or None
        filter = 'filter' in self.request.keys() \
            and self.get_filter_terms() or None
        max = 'max' in self.request.keys() \
            and self.request.get('max') or None
        try:
            max = int(max)
        except:
            max = None

        if self.is_ical_export:
            calendar = self.catalog.calendar(search=search, filter=filter)
            return utils.render_ical_response(self.request, self.context,
                                              calendar)

        elif self.is_json_export:
            export = self.catalog.export(search=search, filter=filter, max=max)
            return utils.render_json_response(self.request, export)

        else:
            return self._template.render(self)

    def update(self, **kwargs):
        daterange = self.request.get('range', self.get_last_daterange())

        # do not trust the user's input blindly
        if not dates.is_valid_daterange(daterange):
            daterange = 'this_month'
        else:
            self.set_last_daterange(daterange)

        state = self.request.get('state', self.get_last_state())

        if not self.show_state_filters or state not in (
            'submitted', 'published', 'archived'
        ):
            state = 'published'
        else:
            self.set_last_state(state)

        self.catalog.state = state
        self.catalog.daterange = daterange

        if not self.is_ical_export and not self.is_json_export:
            super(EventsDirectoryView, self).update(**kwargs)

    def groups(self, items):
        """ Returns the given occurrences grouped by human_date. """
        groups = grouped_occurrences(items, self.request)

        for key, items in groups.items():
            for ix, item in enumerate(items):
                items[ix] = item.get_object()

        return groups

    def translate(self, text, domain="seantis.dir.events"):
        return utils.translate(self.request, text, domain)

    @property
    def show_state_filters(self):
        return getSecurityManager().checkPermission(
            permissions.ReviewPortalContent, self.context
        )

    @cached_property
    def batch(self):
        # use a custom batch whose items are lazy evaluated on __getitem__
        start = int(self.request.get('b_start') or 0)
        lazy_list = self.catalog.lazy_list

        # seantis.dir.events lazy list implementation currently cannot
        # deal with orphans.
        return Batch(lazy_list, directory.ITEMSPERPAGE, start, orphan=0)

    @property
    def selected_state(self):
        return self.catalog.state

    def state_filter_list(self):

        submitted = utils.translate(self.request, _(u'Submitted'))
        submitted += u' (%i)' % self.catalog.submitted_count

        return [
            ('submitted', submitted),
            ('published', _(u'Published'))
        ]

    def state_url(self, method):
        return self.directory.absolute_url() + '?state=' + method

    @utils.webcal
    def ical_url(self, for_all):
        """ Returns the ical url of the current view. """
        url = self.daterange_url('this_year') + '&type=ical'

        if for_all:
            return url

        action, param = self.primary_action()

        if action not in (self.search, self.filter):
            return ''

        if action == self.search:
            if param:
                return url + '&search=true&searchtext=%s' % param
            else:
                return ''

        if action == self.filter:
            terms = dict([(k, v) for k, v in param.items() if v != '!empty'])

            if not terms:
                return ''

            url += '&filter=true'

            for item in terms.items():
                url += '&%s=%s' % item

            return url


class TermsView(grok.View):

    grok.name('terms')
    grok.context(IEventsDirectory)
    grok.require('zope2.View')

    label = _(u'Terms and Conditions')
    template = grok.PageTemplateFile('templates/terms.pt')


class CleanupView(grok.View):

    grok.name('cleanup')
    grok.context(IEventsDirectory)
    grok.require('zope2.View')

    def render(self):

        # dryrun must be disabled explicitly using &run=1
        dryrun = not self.request.get('run') == '1'

        # this maintenance feature may be run unrestricted as it does not
        # leak any information and it's behavior cannot be altered by the
        # user. This allows for easy use via cronjobs.
        execute_under_special_role(
            getSite(), 'Manager',
            maintenance.cleanup_directory, self.context, dryrun
        )

        return u''



from datetime import datetime, timedelta
from DateTime.DateTime import DateTime

from seantis.dir.events.dates import to_utc
from seantis.dir.events.tests import IntegrationTestCase
from seantis.dir.events import maintenance


class TestMaintenance(IntegrationTestCase):

    def setUp(self):
        super(TestMaintenance, self).setUp()
        # login to gain the right to create events
        self.login_admin()

    def tearDown(self):
        super(TestMaintenance, self).tearDown()
        self.logout()

    def test_remove_stale_previews(self):
        preview = self.create_event()
        self.assertEqual(preview.state, 'preview')

        run = lambda: maintenance.remove_stale_previews(
            self.directory, dryrun=True
        )

        self.assertEqual(run(), [])

        preview.modification_date = DateTime(
            datetime.utcnow() - timedelta(days=2, microseconds=1)
        )
        preview.reindexObject(idxs=['modified'])

        self.assertEqual(run(), [preview.id])

        preview.submit()

        self.assertEqual(run(), [])

    def test_archive_past_events(self):

        run = lambda: maintenance.archive_past_events(
            self.directory, dryrun=True
        )

        published = self.create_event()
        self.assertEqual(run(), [])

        published.submit()
        self.assertEqual(run(), [])

        published.publish()
        self.assertEqual(run(), [])

        published.start = to_utc(datetime.utcnow() - timedelta(days=3))
        published.end = to_utc(datetime.utcnow() - timedelta(days=3))
        published.reindexObject(idxs=['start', 'end'])

        self.assertEqual(run(), [published.id])

        published.start += timedelta(days=2)
        published.end += timedelta(days=2)
        published.reindexObject(idxs=['start', 'end'])
        self.assertEqual(run(), [])

        published.start = to_utc(datetime.utcnow() - timedelta(days=3))
        published.end = to_utc(datetime.utcnow() - timedelta(days=3))
        published.reindexObject(idxs=['start', 'end'])
        self.assertEqual(run(), [published.id])

        published.recurrence = 'RRULE:FREQ=WEEKLY;COUNT=10'
        self.assertEqual(run(), [])

        published.start = to_utc(datetime.utcnow() - timedelta(days=100))
        published.end = to_utc(datetime.utcnow() - timedelta(days=100))
        published.recurrence = 'RRULE:FREQ=WEEKLY;COUNT=3'
        published.reindexObject(idxs=['start', 'end'])

        self.assertEqual(run(), [published.id])

    def test_remove_archived_events(self):

        run = lambda: maintenance.remove_archived_events(
            self.directory, dryrun=True
        )

        archived = self.create_event()
        self.assertEqual(run(), [])

        archived.submit()
        self.assertEqual(run(), [])

        archived.publish()
        self.assertEqual(run(), [])

        archived.start = to_utc(datetime.utcnow() - timedelta(days=31))
        archived.end = to_utc(datetime.utcnow() - timedelta(days=31))
        archived.reindexObject(idxs=['start', 'end'])
        self.assertEqual(run(), [])

        archived.archive()
        self.assertEqual(run(), [archived.id])

        archived.start = to_utc(datetime.utcnow() - timedelta(days=10))
        archived.end = to_utc(datetime.utcnow() - timedelta(days=10))
        archived.reindexObject(idxs=['start', 'end'])

        self.assertEqual(run(), [])



import logging
log = logging.getLogger('seantis.dir.events')

import hashlib
import inspect
import transaction
import isodate
import pytz

from functools32 import lru_cache

from datetime import datetime
from urllib import urlopen
from itertools import groupby

from five import grok

from collective.geo.geographer.interfaces import IWriteGeoreferenced
from Products.CMFCore.utils import getToolByName
from plone.namedfile import NamedFile, NamedImage
from plone.dexterity.utils import createContentInContainer
from zope.interface import alsoProvides
from zope.annotation.interfaces import IAnnotations

from seantis.dir.base.interfaces import (
    IDirectoryCatalog,
    IDirectoryCategorized
)
from seantis.dir.base.directory import DirectoryCatalogMixin
from seantis.dir.events.interfaces import (
    IEventsDirectory,
    IExternalEvent
)


signature_method = 'fetch_events'


def available_sources():
    """ Search through the available modules of seantis.dir.events.sources
    and make available the ones which implement the signature_method.

    The result is a dictionary (name of source => function).

    """

    from seantis.dir.events import sources

    callables = {}

    for name, module in inspect.getmembers(sources, inspect.ismodule):
        functions = dict(inspect.getmembers(module, inspect.isfunction))

        if signature_method in functions:
            callables[name] = functions[signature_method]

    return callables


class FetchView(grok.View, DirectoryCatalogMixin):

    grok.name('fetch')
    grok.context(IEventsDirectory)
    grok.require('cmf.ManagePortal')

    @lru_cache(maxsize=50)
    def download(self, url):
        return urlopen(url).read()

    def disable_indexing(self):
        self.context._v_fetching = True

    def enable_indexing(self):
        self.context._v_fetching = False

    def fetch(
        self, source, function, limit=None, reimport=False, source_ids=[]
    ):

        start = datetime.now()
        log.info('begin fetching events for %s' % source)

        self.disable_indexing()

        try:
            imported = self._fetch(
                source, function, limit, reimport, source_ids
            )
        finally:
            self.enable_indexing()

        log.info('reindexing commited events')

        # reindex in the ZCatalog
        for event in imported:
            event.reindexObject()

        self.context.reindexObject()

        # reindex in the Events Catalog
        IDirectoryCatalog(self.context).reindex()

        runtime = datetime.now() - start
        minutes = runtime.total_seconds() // 60
        seconds = runtime.seconds - minutes * 60

        log.info('imported %i events in %i minutes, %i seconds' % (
            len(imported), minutes, seconds
        ))

    def get_last_update_time(self):
        assert self.annotation_key

        isostring = IAnnotations(self.context).get(self.annotation_key, None)

        if not isostring:
            return None
        else:
            return isodate.parse_datetime(isostring)

    def set_last_update_time(self, dt):
        assert self.annotation_key

        assert isinstance(dt, datetime)
        assert dt.tzinfo, "please use a timezone aware datetime"

        # use string to store date to ensure that the annotation
        # doesn't cause problems in the future
        annotations = IAnnotations(self.context)
        annotations[self.annotation_key] = isodate.datetime_isoformat(dt)

    def _fetch(
        self, source, function, limit=None, reimport=False, source_ids=[]
    ):

        events = sorted(
            function(self.context, self.request),
            key=lambda e: (e['last_update'], e['source_id'])
        )

        fetch_ids = set(event['fetch_id'] for event in events)
        assert len(fetch_ids) == 1, """
            Each event needs a fetch_id which describes the id of the
            whole fetch process and is therefore the same for all events
            in a single fetch. See seantis.dir.events.source.guidle:events
        """

        self.annotation_key = hashlib.sha1(list(fetch_ids)[0]).hexdigest()
        last_update = self.get_last_update_time()
        last_update_in_run = datetime.min.replace(tzinfo=pytz.timezone('utc'))

        if not last_update:
            log.info('initial import')
            changed_offers_only = False
        elif reimport:
            log.info('reimport everything')
            changed_offers_only = False
        else:
            log.info('importing updates since {}'.format(last_update))
            changed_offers_only = True

        total = len(events) if not limit else limit
        existing = self.groupby_source_id(self.existing_events(source))

        workflowTool = getToolByName(self.context, 'portal_workflow')

        categories = dict(cat1=set(), cat2=set())
        imported = []

        limit_reached_id = None

        for ix, event in enumerate(events):

            if limit_reached_id and limit_reached_id != event['source_id']:
                break

            if source_ids and event['source_id'] not in source_ids:
                continue

            assert 'last_update' in event, """
                Each event needs a last_update datetime info which is used
                to determine if any changes were done. This is used for
                importing only changed events.
            """

            if last_update_in_run < event['last_update']:
                last_update_in_run = event['last_update']

            if changed_offers_only and event['source_id'] in existing:
                if event['last_update'] <= last_update:
                    log.info('skipping %s @ %s' % (
                        event['title'],
                        event['start'].strftime('%d.%m.%Y %H:%M')
                    ))
                    continue

            # keep a set of all categories for the suggestions
            for cat in categories:
                categories[cat] |= event[cat]

            # for testing
            if limit and len(imported) > limit and not limit_reached_id:
                log.info('reached limit of %i events' % limit)
                # don't quit right away, all events of the same source_id
                # need to be imported first since they have the same
                # update_time
                limit_reached_id = event['source_id']

            # flush to disk every 500th event to keep memory usage low
            if len(imported) % 500 == 0:
                transaction.savepoint(True)

            log.info('importing %i/%i %s @ %s' % (
                len(imported), total, event['title'],
                event['start'].strftime('%d.%m.%Y %H:%M')
            ))

            event['source'] = source

            # source id's are not necessarily unique as a single external
            # event might have to be represented as more than one event in
            # seantis.dir.events - therefore updating is done through
            # deleting first, adding second

            if event['source_id'] in existing:
                for e in existing[event['source_id']]:
                    self.context._delObject(e.id, suppress_events=True)
                del existing[event['source_id']]

            # image and attachments are downloaded
            downloads = {
                'image': NamedImage,
                'attachment_1': NamedFile,
                'attachment_2': NamedFile
            }

            def allow_download(download, url):

                if download != 'image':
                    return True

                # whitelist the images that are known to work
                # not working is *.bmp. We could convert but I'd rather
                # force people to use a sane format
                return url.lower().endswith(('png', 'jpg', 'jpeg'))

            for download, method in downloads.items():
                url = event.get(download)

                if not url or not allow_download(download, url):
                    event[download] = None
                else:
                    event[download] = method(self.download(url))

            # latitude and longitude are set through the interface
            lat, lon = event.get('latitude'), event.get('longitude')

            if lat:
                del event['latitude']

            if lon:
                del event['longitude']

            # so are categories
            cats = map(event.get, ('cat1', 'cat2'))
            del event['cat1']
            del event['cat2']

            assert 'cat3' not in event and 'cat4' not in event, """
                unsupported categories
            """

            obj = createContentInContainer(
                self.context, 'seantis.dir.events.item',
                checkConstraints=False,
                **event
            )

            # set coordinates now
            if lat and lon:
                IWriteGeoreferenced(obj).setGeoInterface(
                    'Point', map(float, (lon, lat))
                )

            # followed by the categories
            IDirectoryCategorized(obj).cat1 = list(cats[0])
            IDirectoryCategorized(obj).cat2 = list(cats[1])

            workflowTool.doActionFor(obj, 'submit')
            workflowTool.doActionFor(obj, 'publish')

            for download in downloads:
                getattr(obj, download)

            alsoProvides(obj, IExternalEvent)
            imported.append(obj)

        self.set_last_update_time(last_update_in_run)

        # add categories to suggestions
        for category in categories:
            key = '%s_suggestions' % category
            existing = set(getattr(self.context, key))
            new = categories[category] | existing

            setattr(self.context, key, sorted(new))

        log.info('committing events for %s' % source)
        transaction.commit()

        return imported

    def existing_events(self, source):

        catalog = getToolByName(self.context, 'portal_catalog')
        candidates = catalog(object_provides=IExternalEvent.__identifier__)

        events = []
        for obj in (c.getObject() for c in candidates):
            if obj.source == source:
                events.append(obj)

        return events

    def groupby_source_id(self, events):

        ids = {}

        keyfn = lambda e: e.source_id
        events.sort(key=keyfn)

        for key, items in groupby(events, lambda e: e.source_id):
            ids[key] = [e for e in items]

        return ids

    def render(self):
        source = self.request.get('source')
        sources = available_sources()

        assert source in sources

        limit = int(self.request.get('limit', 0))
        reimport = bool(self.request.get('reimport', False))
        ids = self.request.get('source-ids', '').split(',')

        self.source = source
        self.fetch(
            source, sources[source], limit, reimport, all(ids) and ids or None
        )

        IDirectoryCatalog(self.context).reindex()



""" Oh boy.

Event Calendars of small municiaplities have a somewhat higher chance of
being empty. So those towns end up sharing their events with other, bigger
towns in the area, increasing their chances of having a reasonably
full calendar.

This is a perfectly fine approach, but it has one drawback. There's only
one calendar with one corporate design. The 'CD'? Not a holy cow that
a mere mortal may slaughter.

To enable this, the idea was to give small towns the ability to use a
custom url to which they can link on their sites. Someone entering
the event calendar through that url would then be guaranteed to see
an alternative design throughout his stay.

So we ended up implementing what awaits below. We allow customers to
use special urls on event directories. We interpret those urls, get
custom properties from the directory and rewrite all the urls and
redirects having to do with said directory.

For the future we very much want to implement a different system to avoid
empty calendars, maybe using P2P calendar syncing. But that will take
a lot of work, so for now that's that.

"""

import re
import json

import logging
log = logging.getLogger('seantis.dir.events')

from urlparse import urlparse, urlunparse
from urllib import quote
from five import grok

from zope.annotation.interfaces import IAnnotations
from zope.component import adapts
from zope.component.hooks import getSite
from zope.interface import implements
from zope.publisher.interfaces import IPublishTraverse
from zope.publisher.interfaces.browser import IBrowserRequest
from zExceptions import NotFound
from ZPublisher.BaseRequest import DefaultPublishTraverse
from ZPublisher.interfaces import IPubSuccess

from plone.app.layout.viewlets.interfaces import IHtmlHeadLinks
from plone.app.theming.transform import ThemeTransform
from plone.transformchain.interfaces import ITransform

from seantis.dir.base.utils import cached_property
from seantis.dir.base.directory import DirectoryCatalogMixin
from seantis.dir.base.interfaces import IDirectoryRoot, IDirectoryPage
from seantis.dir.events.interfaces import IEventsDirectory

# keys for annotation
key_pageid = 'seantis.dir.events.pageid'
key_url = 'seantis.dir.events.directory_url'

# url path matching pattern
pattern = r'[~-]{1}([a-zA-Z]+)[-]?.*'


def pageid_from_name(string):
    """ Extract the pageid from a path name
    e.g. 'test' in 'https://host/test/asdf

    Return None if unsuccessful.

    """
    if not string:
        return None

    # sometimes the token ends up being quoted, even though '~' is a
    # valid character in the path segment of an url
    string = string.replace(quote('~'), '~')
    match = re.match(pattern, string)

    return match.group(1) if match else None


def urlreplace(url, **kwargs):

    parts = urlparse(url)._asdict()

    for key, value in kwargs.items():
        if callable(value):
            parts[key] = value(parts[key])
        else:
            parts[key] = kwargs[value]

    return urlunparse((
        parts['scheme'],
        parts['netloc'],
        parts['path'],
        parts['params'],
        parts['query'],
        parts['fragment']
    ))


class CustomPageRequest(object):

    """ Annotates the given request with everything that's needed to
    serve custom css / styles & content and offers methods to that effect.

    """

    def __init__(self, request):
        self.request = request
        self.annotations = IAnnotations(request)

    def hook(self, pageid, directory):
        """ Hook the request with the given pageid and the directory url. """

        self.pageid = pageid
        self.url = directory.absolute_url()

    @cached_property
    def apply_transformation(self):
        """ Return true if the urls should be transformed. """
        return bool(self.pageid and self.url)

    # the page id is used as a key to look up properties on the directory
    def get_pageid(self):
        return self.annotations.get(key_pageid)

    def set_pageid(self, pageid):
        self.annotations[key_pageid] = pageid

    pageid = property(get_pageid, set_pageid)

    # the url defines the original url of the directory which must be
    # replaced in the response body and in redirect locations
    def get_url(self):
        return self.annotations.get(key_url)

    def set_url(self, url):
        self.annotations[key_url] = url

    url = property(get_url, set_url)

    @cached_property
    def original_path(self):
        path = urlparse(self.url).path
        return path.endswith('/') and path[:-1] or path

    @cached_property
    def replacement_path(self):
        return self.original_path + '/' + self.token

    @cached_property
    def token(self):
        return '~' + self.pageid

    @cached_property
    def quoted_token(self):
        return quote(self.token)

    def transform_url(self, url):
        """ Transform the given url. """

        # if the url is already transformed, don't
        if self.token in url or self.quoted_token in url:
            return url

        return urlreplace(
            url,
            path=lambda p: p.replace(
                self.original_path, self.replacement_path, 1
            )
        )

    def properties(self, directory):
        """ Return the custom properties by page id for the given
        directory.

        """
        if not self.pageid:
            return {}

        if not hasattr(directory, self.pageid):
            return {}

        try:
            return json.loads(getattr(directory, self.pageid))
        except:
            log.exception('pageid json for %s is mal-formed' % self.pageid)

            return {}

    def custom_directory(self, directory):
        """ Return the custom directory from which title, description and
        image are loaded if a page id is found in the request.

        Returns the given directory if the custom directory cannot be found.

        """
        props = self.properties(directory)
        path = props.get('directory')

        if not path:
            return directory

        obj = getSite().unrestrictedTraverse(str(path))
        if IEventsDirectory.providedBy(obj):
            return obj
        else:
            return directory

    def custom_file(self, directory, type):
        """ Return the either the custom css or custom js file. """
        assert type in ('css', 'js')

        path = self.properties(directory).get(type)
        if not path:
            return None

        try:
            return getSite().unrestrictedTraverse(str(path))
        except NotFound:
            return None


class CustomDirectory(object):
    """ Mixin to retrieve the custom directory in instances that have the
    directory as context.

    """

    @cached_property
    def custom_directory(self):
        pagerequest = CustomPageRequest(self.request)
        return pagerequest.custom_directory(self.context)


class CustomPageHook(object):
    """ Intercepts traversal of event directories and sets up the custom
    page request if a custom token is found during traversal.

    e.g.

    'events' is the directory in the following urls:

    http://host/site/events/~custom/ => hook 'custom'
    http://host/site/events/event/~town/edit => hook 'town'
    http://host/site/~town/events/ => no hook as it is no sub-path

    """

    implements(IPublishTraverse)

    def publishTraverse(self, request, name):

        try:

            pageid = pageid_from_name(name)
            if pageid:

                pagerequest = CustomPageRequest(request)
                pagerequest.hook(pageid, self)

                return self

        except Exception:
            # Zope will consume this exception and ignore it, so
            # we log it for the record
            log.exception('error while traversing')
            raise

        return DefaultPublishTraverse(self, request).publishTraverse(
            request, name
        )


class CustomPageViewlet(grok.Viewlet, DirectoryCatalogMixin):
    """ Viewlet which renders the custom css and javascript links in the
    header. """

    grok.context(IDirectoryRoot)
    grok.name('seantis.dir.events.custom')
    grok.require('zope2.View')
    grok.viewletmanager(IHtmlHeadLinks)

    grok.order(99)  # as far below as possible

    template = grok.PageTemplateFile('templates/custom.pt')

    @cached_property
    def pagerequest(self):
        return CustomPageRequest(self.request)

    def custom_css(self):
        f = self.pagerequest.custom_file(self.context, 'css')
        return f and f.data or ''

    def custom_script(self):
        f = self.pagerequest.custom_file(self.context, 'js')
        return f and f.data or ''


class URLTransform(object):
    """ Transforms the response body, rewriting the a.href and form.action
    to include the custom page-id.

    """

    implements(ITransform)
    adapts(IDirectoryPage, IBrowserRequest)

    order = 9001

    def __init__(self, published, request):
        self.published = published
        self.request = request

    @cached_property
    def pagerequest(self):
        return CustomPageRequest(self.request)

    def transformString(self, result, encoding):
        return self.transformIterable([result], encoding)

    def transformUnicode(self, result, encoding):
        return self.transformIterable([result], encoding)

    def transformIterable(self, result, encoding):
        pagerequest = CustomPageRequest(self.request)

        if not pagerequest.apply_transformation:
            return result

        # use the tree from plone.app.theming to avoid building it ourselves
        tree = ThemeTransform(self.published, self.request).parseTree(result)
        tree = tree.tree if tree else None

        if not tree:
            return result

        # save the tree on the instance for testing, when running for real
        # plone.app.theming / diazo will take care of serializing the tree
        self._tree = tree

        path = pagerequest.original_path

        for a in tree.xpath("//a[contains(@href, '%s')]" % path):
            a.attrib['href'] = pagerequest.transform_url(
                a.attrib.get('href', '')
            )

        for form in tree.xpath("//form[contains(@action, '%s')]" % path):
            form.attrib['action'] = pagerequest.transform_url(
                form.attrib.get('action', '')
            )

        return result


@grok.subscribe(IPubSuccess)
def pub_success(e):
    """ Intercept redirects and change the location to include the custom
    page-id.

    """
    response = e.request.response
    if response.status in (301, 302, 303, 307):

        pagerequest = CustomPageRequest(e.request)
        if pagerequest.apply_transformation:
            response.setHeader(
                'Location',
                pagerequest.transform_url(response.getHeader('Location'))
            )



from zope.publisher.browser import TestRequest
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.interface import directlyProvides

from repoze.xmliter.serializer import XMLSerializer

from seantis.dir.events.tests import IntegrationTestCase
from seantis.dir.events import pages


class PagesTestCase(IntegrationTestCase):

    def test_stringmatching(self):
        match = pages.pageid_from_name

        # do not support urls
        self.assertEqual(match('https://google.com'), None)
        self.assertEqual(match('https://google.com/test'), None)
        self.assertEqual(match('https://host/~test'), None)
        self.assertEqual(match('https://host/-test-'), None)
        self.assertEqual(match('https://host/asdf-asdf'), None)

        self.assertEqual(match('~id'), 'id')
        self.assertEqual(match('-id-'), 'id')
        self.assertEqual(match(None), None)
        self.assertEqual(match(''), None)
        self.assertEqual(match('~/-test'), None)
        self.assertEqual(match('~~~asdf'), None)

        # do not support the same characters in the middle
        # as that might catch a proper named object
        self.assertEqual(match('my-folder'), None)
        self.assertEqual(match('folder-1'), None)
        self.assertEqual(match('my~home'), None)

    def test_urltransform(self):
        request = TestRequest()
        directlyProvides(request, IAttributeAnnotatable)

        request.response.setHeader('Content-Type', 'text/html')

        custom = pages.CustomPageRequest(request)
        custom.hook('custompage', self.directory)

        transform = pages.URLTransform(self.directory, request)

        def run(s):
            transform.transformString(s, 'ascii')

            if hasattr(transform, '_tree'):
                return str(XMLSerializer(transform._tree))

            return None

        html = """
        <html>
            <head></head>
            <body>
                <a href="%(url)s"></a>
                <a href="#"></a>
                <form action="%(url)s"></form>
                <form></form>
            </body>
        </html>
        """ % dict(url=self.directory.absolute_url())

        result = run(html)

        url = self.directory.absolute_url() + '/~custompage'
        self.assertEqual(result.count(url), 2)

        replaced = '"%s"' % self.directory.absolute_url()
        self.assertEqual(result.count(replaced), 0)



from datetime import datetime, timedelta

from logging import getLogger
log = getLogger('seantis.dir.events')

from seantis.dir.base.interfaces import IDirectoryCatalog

from seantis.dir.events.dates import to_utc
from seantis.dir.events.interfaces import IEventsDirectoryItem
from seantis.dir.events.recurrence import has_future_occurrences


def remove_stale_previews(directory, dryrun=False):

    catalog = IDirectoryCatalog(directory)
    query = catalog.catalog.unrestrictedSearchResults

    log.info('searching for stale previews')

    past = to_utc(datetime.utcnow() - timedelta(days=2))
    stale_previews = query(
        path={'query': directory.getPhysicalPath(), 'depth': 2},
        object_provides=IEventsDirectoryItem.__identifier__,
        review_state=('preview'),
        modified={'query': past, 'range': 'max'}
    )
    stale_previews = [p.id for p in stale_previews]

    if stale_previews:
        log.info('deleting stale previews -> %s' % str(stale_previews))
        if not dryrun:
            directory.manage_delObjects(stale_previews)
    else:
        log.info('no stale previews found')

    return stale_previews


def archive_past_events(directory, dryrun=False):

    catalog = IDirectoryCatalog(directory)
    query = catalog.catalog

    log.info('archiving past events')

    # events are in the past if they have been over for two days
    # (not one, to ensure that they are really over in all timezones)
    past = to_utc(datetime.utcnow() - timedelta(days=2))
    published_events = query(
        path={'query': directory.getPhysicalPath(), 'depth': 2},
        object_provides=IEventsDirectoryItem.__identifier__,
        review_state=('published', ),
        start={'query': past, 'range': 'max'},
        end={'query': past, 'range': 'max'}
    )

    past_events = []

    for event in published_events:
        event = event.getObject()

        assert event.start < past
        assert event.end < past

        # recurring events may be in the past with one of
        # their occurrences in the future
        if not has_future_occurrences(event, past):
            past_events.append(event)

    ids = [p.id for p in past_events]

    if past_events:
        log.info('archiving past events -> %s' % str(ids))

        if not dryrun:
            for event in past_events:
                event.archive()
    else:
        log.info('no past events found')

    return ids


def remove_archived_events(directory, dryrun=False):

    catalog = IDirectoryCatalog(directory)
    query = catalog.catalog

    log.info('removing archived events')

    past = datetime.utcnow() - timedelta(days=30)
    archived_events = query(
        path={'query': directory.getPhysicalPath(), 'depth': 2},
        object_provides=IEventsDirectoryItem.__identifier__,
        review_state=('archived', ),
        start={'query': past, 'range': 'max'},
        end={'query': past, 'range': 'max'}
    )
    archived_events = [e.id for e in archived_events]

    if archived_events:
        log.info('removing archived events -> %s' % str(archived_events))

        if not dryrun:
            directory.manage_delObjects(archived_events)
    else:
        log.info('no archived events to remove')

    return archived_events


def cleanup_directory(directory, dryrun=True):

    if dryrun:
        log.info('starting dry run cleanup on %s' % directory.absolute_url())
    else:
        log.info('starting real cleanup on %s' % directory.absolute_url())

    remove_stale_previews(directory, dryrun)
    archive_past_events(directory, dryrun)
    remove_archived_events(directory, dryrun)

    log.info('finished cleanup on %s' % directory.absolute_url())
